#include "KMotionDef.h"
#include "Triac.h"

// Services Tool Changer using non-blocking State Machine approach
// The Service routines maintain their state so they can always return
// immediately and later resume from their previous state.

// T O O L   C H A N G E R   S E Q U E N C E
//
// define state that the tool changer may be in

//#define TOOL_CLAMP_BIT 20 	// IO bit number to clamp turret
//#define TOOL_DIST 4250.0	// Steps/counts to move turret to next tool
//#define CLAMP_TIME 1.5    	// seconds to wait for the clamp/unclamp

#define TOOLPOCKETS 6		// Number of tool pockets
//#define TURRET_AXIS 2     	// axis channel for the turret motor
#define CRSLIN_DELAY 0.5	// seconds to delay after carousel in switch triggered
#define CRSLDN_DELAY 0.5	// seconds to delay after carousel down switch triggered
#define DRBUNCLAMP_DELAY 0.5	// seconds to wait for unclamp
#define CRSLSTOP_DELAY	0.04	// Seconds to wait before stopping carousel
#define CRSLUP_DELAY 0.5	// seconds to delay after carousel up switch triggered
#define DRBCLAMP_DELAY 0.5	// seconds to delay after drawbar clamped
#define CRSLOUT_DELAY 0.5	// seconds to delay after carousel out switch triggered
#define CRSLIDX_DEBOUNCE 0.3	//

//Time Out constants
#define TO_CRSLIN 5		// Carousel moving in
#define TO_CRSLIDXSTART 0.8	// Time to allow for carousel to start moving
#define TO_CRSLIDX 1.3		// Time to allow for carousel rotating between index flats
#define TO_CRSLIDXFLAT 0.5	// Time to allow for carousel to remain on the flat

int *ChangerState 	= &persist.UserData[TOOL_STATE_VAR];
int *Tool     		= &persist.UserData[TOOL_VAR];
int *LastTool 		= &persist.UserData[LAST_TOOL_VAR];
int *CrslPos		= &persist.UserData[CRSL_POS];	// divide by 4 to get actual position

double ToolTime;	// used for non blocking time delays
double TimeOutTime;	// used for timing out should something go wrong
double PrintTime;
int ToolPockets;	// used to store number of tool pockets still to move
int CrslDir;		// use to remember what direction to turn carousel
int IndexState = T_IDX_HIGH;		// use this to track if we're waiting for index to go high or low
int TWait = FALSE;	// used for time delay purposes

// Services Tool Change Sequence
void ServiceToolChange(void)
{
	if(Time_sec() > PrintTime){
		//printf(" ChangerState=%d\n",*ChangerState);
		//printf("Current Tool=%d, toolpockets=%d\n",*Tool, ToolPockets);
		
		PrintTime = Time_sec() + 2;
	}
	switch (*ChangerState)
	{
	case T_ESTOP:
	// if we're here, then we've had an e-stop while TC was not idle, so we want to hold TC position, but stop all motor movement
		{
			ClearBit(CRSLRUN);	// stop movement
			ClearBit(CRSLREV);
			*ChangerState = T_FAILED;	// then set status to failed, ready for recovery
			break;
		}
	case T_IDLE:
		{
			break;
		}
	case T_MOVEZ:
	// Moves Z to tool change height
		{
			if(!TWait){
				if(!ReadBit(ZHOMED)){
					printf("Z not homed)\n");
					MsgBoxNoWait("Z not homed. Home Z and re-try tool change.",MB_OK|MB_ICONEXCLAMATION);
					*ChangerState = T_FAILED;
				}
				printf("Move Z to toolchange height\n");
				Move(Z,TCZHEIGHT);	// Move Z to required height
				TWait = TRUE;
			} else {
				if(CheckDone(Z)) {
					*ChangerState = T_PAUSED;
					TWait = FALSE;
				}
			}
			break;
		}
	case T_CRSLIN:
	// Moves Carousel In
		{
			if(!TWait){
				if(ReadBit(CRSLUP) && (*CrslPos != (*LastTool*4))){	// if carousel tool in spindle and position does not match tool,
					printf("Carousel not on position\n");		// abort tool change
					*ChangerState = T_FAILED;
					MsgBoxNoWait("Carousel not on correct slot to move in",MB_OK|MB_ICONEXCLAMATION);
					break;
				} else {		// else carousel is not up, or we're on the correct position
					SetBit(CRSLINR);	// so we can safely move carousel in
				}
				if(ReadBit(CRSLIN))			// is carousel in yet?
				{
					printf("Carousel moved In\n");
					ToolTime = Time_sec() + CRSLIN_DELAY;  // wait until this time
					TWait = TRUE;
				}
			} else {
				if (Time_sec() > ToolTime){
					*ChangerState = T_PAUSED;
					TWait = FALSE;
				}
			}
			break;
		}
	case T_DBUNCLAMP:
	// Unclamps Drawbar
		{
			if(!TWait){
				SetBit(DRBR);
				ToolTime = Time_sec() + DRBUNCLAMP_DELAY;
				printf("Drawbar Unclamping\n");
				TWait = TRUE;
			} else {
				if (Time_sec() > ToolTime)
				{
					*ChangerState = T_PAUSED;
					TWait = FALSE;
				}
			}
			break;
		}
	case T_CRSLDOWN:
	// Moves carousel down
		{
			if(!TWait){
				SetBit(CRSLDWNR);
				if(ReadBit(CRSLDWN))
				{
					printf("Carousel moving Down\n");
					ToolTime = Time_sec() + CRSLDN_DELAY;  // wait until this time
					TWait = TRUE;
				}
			} else {
				if (Time_sec() > ToolTime){
					*ChangerState = T_PAUSED;
					TWait = FALSE;
				}
			}
			break;
		}
	case T_CRSLROTATE:
	// calculates required carousel movement
		{
			if((*Tool*4) == *CrslPos)	// if carousel is already at requested tool position
			{
				printf("Carousel already on position\n");
				*ChangerState = T_PAUSED;	// we simply go back to paused status
			} else {
				// compute shortest rotation direction
				ToolPockets = ((*Tool * 4) - *CrslPos + (TOOLPOCKETS*4)) % (TOOLPOCKETS*4); // this gives us the number of quarters between tools
				printf("Carousel Position = %d\n", *CrslPos);
				printf("Tool requested = %d\n", *Tool);
				printf("ToolPockets pre direction = %d\n", ToolPockets);
				if((ToolPockets % 2) == 1 ) {		// if we have ended up with a result that is not a half (one geneva rotation) step between tools
					*ChangerState = T_FAILED;		// exit with a fail
					MsgBoxNoWait("Carousel rotation calculation failure (partial rotation result).",MB_OK|MB_ICONEXCLAMATION);
				} else {
					
					if(ToolPockets < ((TOOLPOCKETS*4) / 2))
					{
						CrslDir = T_CW;
						IndexState = T_IDX_HIGH;
						//ToolPockets = ToolPockets*2;	// multiply tool positions by 2, as we need 2 revolutions per tool pocket
						printf("Rotate CW %d steps, from %d to %d\n", ToolPockets, *LastTool, *Tool);
						TimeOutTime = Time_sec() + TO_CRSLIDXSTART;
						*ChangerState = T_CRSLRUN;
					} else {
						CrslDir = T_CCW;
						IndexState = T_IDX_HIGH;
						ToolPockets = ((TOOLPOCKETS*4) - ToolPockets);
						printf("Rotate CCW %d steps, from %d to %d\n", ToolPockets, *LastTool, *Tool);
						TimeOutTime = Time_sec() + TO_CRSLIDXSTART;
						*ChangerState = T_CRSLRUN;
					}
				}
			}
			break;
		}
	case T_CRSLRUN:
	// rotates carousel
		{
			if(TimeOutTime < Time_sec()){	// if we've timed out, stop rotation,
				ClearBit(CRSLRUN);	// stop movement
				ClearBit(CRSLREV);
				ClearBit(TCINITOK);	// deinitialise TC, as we'll of lost carousel position
				*ChangerState = T_FAILED;	// Fail tool change
				printf("Timeout\n");
				printf("Time_sec=%f\n",Time_sec());
				MsgBoxNoWait("Carousel Rotation time out (carousel jammed, or index sensor failure)",MB_OK|MB_ICONEXCLAMATION); // and finally warn user
				break;
			}
	
			if(!TWait)
			{
				if(CrslDir == T_CW) SetBit(CRSLREV);	// if we need CCW, enable CCW relay
				SetBit(CRSLRUN);				// and enable run relay
				if(IndexState == T_IDX_HIGH && (Time_sec() > ToolTime))	// wait for index to go high (i.e reach flat)
				{
					if(ReadBit(CRSLIDX)){		// when index goes high - we're of the flat
						if(CrslDir == T_CCW){		// we're decreasing carousel position
							*CrslPos = *CrslPos - 1;
						} else {					// else we're increasing carousel position
							if(*CrslPos == 24){		// is we're at 24
								*CrslPos = 1;	// we need to roll over to 1
							} else {
								*CrslPos = *CrslPos + 1;
							}
						}
						printf("Carousel Position=%d\n",*CrslPos);
						printf("Index High\n");
						printf("Time_sec=%f\n",Time_sec());
						ToolTime = Time_sec() + CRSLIDX_DEBOUNCE;
						TimeOutTime = Time_sec() + TO_CRSLIDX;	// set the timeout (each rotation should take under ~1 sec, so if we exceed the timeout, we've got a problem)
						printf("TimeOutTime set - %f\n", TimeOutTime);
						IndexState = T_IDX_LOW;		// we now need to wait for index to go high
					}
				}
				else if(IndexState == T_IDX_LOW)
				{
					if(!ReadBit(CRSLIDX) && (Time_sec() > ToolTime))
					{	// when index goes low
						if(CrslDir == T_CCW){		// we're decreasing carousel position
							if(*CrslPos == 1){	// if we're at 1
								*CrslPos = 24;		// we need to roll over to 24
							} else {
								*CrslPos = *CrslPos - 1;
							}
						} else {					// else we're increasing carousel position
							*CrslPos = *CrslPos + 1;
						}
						printf("Carousel Position=%d\n",*CrslPos);
						printf("Index Low\n");
						printf("Time_sec=%f\n",Time_sec());
						ToolTime = Time_sec() + CRSLIDX_DEBOUNCE;
						TimeOutTime = Time_sec() + TO_CRSLIDXFLAT;	// set the timeout (each rotation should take under ~1 sec, so if we exceed the timeout, we've got a problem)
						printf("TimeOutTime set - %f\n", TimeOutTime);
						if(ToolPockets > 2)
						{	// if we still need to move
							ToolPockets = ToolPockets - 2;		// reduce counter by one
							printf("ToolPockets=%d\n",ToolPockets);
							ToolTime = Time_sec() + CRSLIDX_DEBOUNCE;
							IndexState = T_IDX_HIGH;	// and reset state so we continue
						} else {	// else we've reached the required position
							printf("ToolPockets=%d\n",ToolPockets);
							ToolTime = Time_sec() + CRSLSTOP_DELAY;	// so set stop time delay
							TWait = TRUE;				// Set wait flag so we now exit this main loop and stop things
						}
					}
				}
			} else {
				if (Time_sec() > ToolTime){
					ClearBit(CRSLRUN);		// once timer elapsed, disable all rotation
					ClearBit(CRSLREV);
					*ChangerState = T_PAUSED;
					TWait = FALSE;
				}
			}
			
			break;
		}
	case T_CRSLUP:
		{
			if(!TWait){
				ClearBit(CRSLDWNR);
				if(ReadBit(CRSLUP))
				{
					printf("Carousel moving Up\n");
					ToolTime = Time_sec() + CRSLUP_DELAY;  // wait until this time
					TWait = TRUE;
				}
			} else {
				if (Time_sec() > ToolTime){
					*ChangerState = T_PAUSED;
					TWait = FALSE;
				}
			}
			break;
		}
	case T_DBCLAMP:
		{
			if(!TWait){
				ClearBit(DRBR);
				ToolTime = Time_sec() + DRBCLAMP_DELAY;
				printf("Drawbar Clamping\n");
				TWait = TRUE;
			} else {
				if (Time_sec() > ToolTime)
				{
					*ChangerState = T_PAUSED;
					TWait = FALSE;
				}
			}
			break;
		}
	case T_CRSLOUT:
		{
			if(!TWait){
				ClearBit(CRSLINR);
				if(ReadBit(CRSLOUT))
				{
					printf("Carousel moving Out\n");
					ToolTime = Time_sec() + CRSLOUT_DELAY;  // wait until this time
					TWait = TRUE;
				}
			} else {
				if (Time_sec() > ToolTime){
					*ChangerState = T_PAUSED;
					TWait = FALSE;
				}
			}
			break;
		}
	case T_END:
		{
			if(Time_sec() > ToolTime)
			{
				printf("Tool Change Complete\n");
				*LastTool = *Tool;  // remember where we are
				*ChangerState = T_IDLE;
			}
			break;
		}
	}
}

// Loop servicing things
main()
{
	for (;;)  // loop forever
	{
		WaitNextTimeSlice();
		ServiceToolChange();
	}
}
